@@ -1,7 +1,7 @@
 from __future__ import division
 
 import abc
-
+import collections
 import base64
 import functools
 import imp
@@ -44,16 +44,14 @@ import cloudpickle
 from cloudpickle.cloudpickle import _find_module, _make_empty_cell, cell_set
 
 from .testutils import subprocess_pickle_echo
+from .testutils import assert_run_python_script
 
 
-HAVE_WEAKSET = hasattr(weakref, 'WeakSet')
-
-
-def pickle_depickle(obj):
+def pickle_depickle(obj, protocol=cloudpickle.DEFAULT_PROTOCOL):
     """Helper function to test whether object pickled with cloudpickle can be
     depickled with pickle
     """
-    return pickle.loads(cloudpickle.dumps(obj))
+    return pickle.loads(cloudpickle.dumps(obj, protocol=protocol))
 
 
 class CloudPicklerTest(unittest.TestCase):
@@ -64,15 +62,17 @@ class CloudPicklerTest(unittest.TestCase):
 
 class CloudPickleTest(unittest.TestCase):
 
+    protocol = cloudpickle.DEFAULT_PROTOCOL
+
     def test_itemgetter(self):
         d = range(10)
         getter = itemgetter(1)
 
-        getter2 = pickle_depickle(getter)
+        getter2 = pickle_depickle(getter, protocol=self.protocol)
         self.assertEqual(getter(d), getter2(d))
 
         getter = itemgetter(0, 3)
-        getter2 = pickle_depickle(getter)
+        getter2 = pickle_depickle(getter, protocol=self.protocol)
         self.assertEqual(getter(d), getter2(d))
 
     def test_attrgetter(self):
@@ -81,18 +81,18 @@ class CloudPickleTest(unittest.TestCase):
                 return item
         d = C()
         getter = attrgetter("a")
-        getter2 = pickle_depickle(getter)
+        getter2 = pickle_depickle(getter, protocol=self.protocol)
         self.assertEqual(getter(d), getter2(d))
         getter = attrgetter("a", "b")
-        getter2 = pickle_depickle(getter)
+        getter2 = pickle_depickle(getter, protocol=self.protocol)
         self.assertEqual(getter(d), getter2(d))
 
         d.e = C()
         getter = attrgetter("e.a")
-        getter2 = pickle_depickle(getter)
+        getter2 = pickle_depickle(getter, protocol=self.protocol)
         self.assertEqual(getter(d), getter2(d))
         getter = attrgetter("e.a", "e.b")
-        getter2 = pickle_depickle(getter)
+        getter2 = pickle_depickle(getter, protocol=self.protocol)
         self.assertEqual(getter(d), getter2(d))
 
     # Regression test for SPARK-3415
@@ -124,12 +124,32 @@ class CloudPickleTest(unittest.TestCase):
     def test_buffer(self):
         try:
             buffer_obj = buffer("Hello")
-            self.assertEqual(pickle_depickle(buffer_obj), str(buffer_obj))
+            buffer_clone = pickle_depickle(buffer_obj, protocol=self.protocol)
+            self.assertEqual(buffer_clone, str(buffer_obj))
             buffer_obj = buffer("Hello", 2, 3)
-            self.assertEqual(pickle_depickle(buffer_obj), str(buffer_obj))
+            buffer_clone = pickle_depickle(buffer_obj, protocol=self.protocol)
+            self.assertEqual(buffer_clone, str(buffer_obj))
         except NameError:  # Python 3 does no longer support buffers
             pass
 
+    def test_memoryview(self):
+        buffer_obj = memoryview(b"Hello")
+        self.assertEqual(pickle_depickle(buffer_obj, protocol=self.protocol),
+                         buffer_obj.tobytes())
+
+    @pytest.mark.skipif(sys.version_info < (3, 4),
+                        reason="non-contiguous memoryview not implemented in "
+                               "old Python versions")
+    def test_sliced_and_non_contiguous_memoryview(self):
+        buffer_obj = memoryview(b"Hello!" * 3)[2:15:2]
+        self.assertEqual(pickle_depickle(buffer_obj, protocol=self.protocol),
+                         buffer_obj.tobytes())
+
+    def test_large_memoryview(self):
+        buffer_obj = memoryview(b"Hello!" * int(1e7))
+        self.assertEqual(pickle_depickle(buffer_obj, protocol=self.protocol),
+                         buffer_obj.tobytes())
+
     def test_lambda(self):
         self.assertEqual(pickle_depickle(lambda: 1)(), 1)
 
@@ -137,7 +157,7 @@ class CloudPickleTest(unittest.TestCase):
         a, b = 1, 2
         f1 = lambda x: x + a
         f2 = lambda x: f1(x) // b
-        self.assertEqual(pickle_depickle(f2)(1), 1)
+        self.assertEqual(pickle_depickle(f2, protocol=self.protocol)(1), 1)
 
     def test_recursive_closure(self):
         def f1():
@@ -166,7 +186,7 @@ class CloudPickleTest(unittest.TestCase):
             msg='f actually has closure cells!',
         )
 
-        g = pickle_depickle(f)
+        g = pickle_depickle(f, protocol=self.protocol)
 
         self.assertTrue(
             g.__closure__ is None,
@@ -187,7 +207,7 @@ class CloudPickleTest(unittest.TestCase):
         with pytest.raises(NameError):
             g1()
 
-        g2 = pickle_depickle(g1)
+        g2 = pickle_depickle(g1, protocol=self.protocol)
         with pytest.raises(NameError):
             g2()
 
@@ -217,7 +237,7 @@ class CloudPickleTest(unittest.TestCase):
         self.assertEqual(Derived().method(), 2)
 
         # Pickle and unpickle the class.
-        UnpickledDerived = pickle_depickle(Derived)
+        UnpickledDerived = pickle_depickle(Derived, protocol=self.protocol)
         self.assertEqual(UnpickledDerived().method(), 2)
 
         # We have special logic for handling __doc__ because it's a readonly
@@ -226,7 +246,7 @@ class CloudPickleTest(unittest.TestCase):
 
         # Pickle and unpickle an instance.
         orig_d = Derived()
-        d = pickle_depickle(orig_d)
+        d = pickle_depickle(orig_d, protocol=self.protocol)
         self.assertEqual(d.method(), 2)
 
     def test_cycle_in_classdict_globals(self):
@@ -239,7 +259,7 @@ class CloudPickleTest(unittest.TestCase):
         C.C_again = C
         C.instance_of_C = C()
 
-        depickled_C = pickle_depickle(C)
+        depickled_C = pickle_depickle(C, protocol=self.protocol)
         depickled_instance = pickle_depickle(C())
 
         # Test instance of depickled class.
@@ -258,8 +278,8 @@ class CloudPickleTest(unittest.TestCase):
             return (x + y) / LOCAL_CONSTANT
 
         # pickle the function definition
-        self.assertEqual(pickle_depickle(some_function)(41, 1), 1)
-        self.assertEqual(pickle_depickle(some_function)(81, 3), 2)
+        self.assertEqual(pickle_depickle(some_function, protocol=self.protocol)(41, 1), 1)
+        self.assertEqual(pickle_depickle(some_function, protocol=self.protocol)(81, 3), 2)
 
         hidden_constant = lambda: LOCAL_CONSTANT
 
@@ -275,26 +295,30 @@ class CloudPickleTest(unittest.TestCase):
                 return self.one() + some_function(x, 1) + self.value
 
         # pickle the class definition
-        self.assertEqual(pickle_depickle(SomeClass)(1).one(), 1)
-        self.assertEqual(pickle_depickle(SomeClass)(5).some_method(41), 7)
-        new_class = subprocess_pickle_echo(SomeClass)
-        self.assertEqual(new_class(5).some_method(41), 7)
+        clone_class = pickle_depickle(SomeClass, protocol=self.protocol)
+        self.assertEqual(clone_class(1).one(), 1)
+        self.assertEqual(clone_class(5).some_method(41), 7)
+        clone_class = subprocess_pickle_echo(SomeClass, protocol=self.protocol)
+        self.assertEqual(clone_class(5).some_method(41), 7)
 
         # pickle the class instances
         self.assertEqual(pickle_depickle(SomeClass(1)).one(), 1)
         self.assertEqual(pickle_depickle(SomeClass(5)).some_method(41), 7)
-        new_instance = subprocess_pickle_echo(SomeClass(5))
+        new_instance = subprocess_pickle_echo(SomeClass(5),
+                                              protocol=self.protocol)
         self.assertEqual(new_instance.some_method(41), 7)
 
         # pickle the method instances
         self.assertEqual(pickle_depickle(SomeClass(1).one)(), 1)
         self.assertEqual(pickle_depickle(SomeClass(5).some_method)(41), 7)
-        new_method = subprocess_pickle_echo(SomeClass(5).some_method)
+        new_method = subprocess_pickle_echo(SomeClass(5).some_method,
+                                            protocol=self.protocol)
         self.assertEqual(new_method(41), 7)
 
     def test_partial(self):
         partial_obj = functools.partial(min, 1)
-        self.assertEqual(pickle_depickle(partial_obj)(4), 1)
+        partial_clone = pickle_depickle(partial_obj, protocol=self.protocol)
+        self.assertEqual(partial_clone(4), 1)
 
     @pytest.mark.skipif(platform.python_implementation() == 'PyPy',
                         reason="Skip numpy and scipy tests on PyPy")
@@ -342,7 +366,7 @@ class CloudPickleTest(unittest.TestCase):
             for i in range(cnt):
                 yield i
 
-        gen2 = pickle_depickle(some_generator)
+        gen2 = pickle_depickle(some_generator, protocol=self.protocol)
 
         assert type(gen2(3)) == type(some_generator(3))
         assert list(gen2(3)) == list(range(3))
@@ -359,8 +383,8 @@ class CloudPickleTest(unittest.TestCase):
         sm = A.__dict__["test_sm"]
         cm = A.__dict__["test_cm"]
 
-        A.test_sm = pickle_depickle(sm)
-        A.test_cm = pickle_depickle(cm)
+        A.test_sm = pickle_depickle(sm, protocol=self.protocol)
+        A.test_cm = pickle_depickle(cm, protocol=self.protocol)
 
         self.assertEqual(A.test_sm(), "sm")
         self.assertEqual(A.test_cm(), "cm")
@@ -381,7 +405,8 @@ class CloudPickleTest(unittest.TestCase):
         # self.assertEqual(g(F(), 1), 2)  # still fails
 
     def test_module(self):
-        self.assertEqual(pickle, pickle_depickle(pickle))
+        pickle_clone = pickle_depickle(pickle, protocol=self.protocol)
+        self.assertEqual(pickle, pickle_clone)
 
     def test_dynamic_module(self):
         mod = imp.new_module('mod')
@@ -389,11 +414,23 @@ class CloudPickleTest(unittest.TestCase):
         x = 1
         def f(y):
             return x + y
+
+        class Foo:
+            def method(self, x):
+                return f(x)
         '''
         exec(textwrap.dedent(code), mod.__dict__)
-        mod2 = pickle_depickle(mod)
+        mod2 = pickle_depickle(mod, protocol=self.protocol)
         self.assertEqual(mod.x, mod2.x)
         self.assertEqual(mod.f(5), mod2.f(5))
+        self.assertEqual(mod.Foo().method(5), mod2.Foo().method(5))
+
+        if platform.python_implementation() != 'PyPy':
+            # XXX: this fails with excessive recursion on PyPy.
+            mod3 = subprocess_pickle_echo(mod, protocol=self.protocol)
+            self.assertEqual(mod.x, mod3.x)
+            self.assertEqual(mod.f(5), mod3.f(5))
+            self.assertEqual(mod.Foo().method(5), mod3.Foo().method(5))
 
         # Test dynamic modules when imported back are singletons
         mod1, mod2 = pickle_depickle([mod, mod])
@@ -411,20 +448,20 @@ class CloudPickleTest(unittest.TestCase):
             _find_module('valid_module')
 
     def test_Ellipsis(self):
-        self.assertEqual(Ellipsis, pickle_depickle(Ellipsis))
+        self.assertEqual(Ellipsis,
+                         pickle_depickle(Ellipsis, protocol=self.protocol))
 
     def test_NotImplemented(self):
-        self.assertEqual(NotImplemented, pickle_depickle(NotImplemented))
+        ExcClone = pickle_depickle(NotImplemented, protocol=self.protocol)
+        self.assertEqual(NotImplemented, ExcClone)
 
-    @pytest.mark.skipif((3, 0) < sys.version_info < (3, 4),
-                        reason="fails due to pickle behavior in Python 3.0-3.3")
     def test_builtin_function_without_module(self):
         on = object.__new__
-        on_depickled = pickle_depickle(on)
+        on_depickled = pickle_depickle(on, protocol=self.protocol)
         self.assertEqual(type(on_depickled(object)), type(object()))
 
         fi = itertools.chain.from_iterable
-        fi_depickled = pickle_depickle(fi)
+        fi_depickled = pickle_depickle(fi, protocol=self.protocol)
         self.assertEqual(list(fi([[1, 2], [3, 4]])), [1, 2, 3, 4])
 
     @pytest.mark.skipif(tornado is None,
@@ -578,7 +615,7 @@ class CloudPickleTest(unittest.TestCase):
 
     def test_logger(self):
         logger = logging.getLogger('cloudpickle.dummy_test_logger')
-        pickled = pickle_depickle(logger)
+        pickled = pickle_depickle(logger, protocol=self.protocol)
         self.assertTrue(pickled is logger, (pickled, logger))
 
         dumped = cloudpickle.dumps(logger)
@@ -612,8 +649,9 @@ class CloudPickleTest(unittest.TestCase):
             def foo(self):
                 return 'it works!'
 
-        depickled_base = pickle_depickle(AbstractClass)
-        depickled_class = pickle_depickle(ConcreteClass)
+        depickled_base = pickle_depickle(AbstractClass, protocol=self.protocol)
+        depickled_class = pickle_depickle(ConcreteClass,
+                                          protocol=self.protocol)
         depickled_instance = pickle_depickle(ConcreteClass())
 
         self.assertEqual(depickled_class().foo(), 'it works!')
@@ -628,7 +666,6 @@ class CloudPickleTest(unittest.TestCase):
 
         self.assertEqual(DepickledBaseSubclass().foo(), 'it works for realz!')
 
-    @pytest.mark.skipif(not HAVE_WEAKSET, reason="WeakSet doesn't exist")
     def test_weakset_identity_preservation(self):
         # Test that weaksets don't lose all their inhabitants if they're
         # pickled in a larger data structure that includes other references to
@@ -641,7 +678,7 @@ class CloudPickleTest(unittest.TestCase):
         obj1, obj2, obj3 = SomeClass(1), SomeClass(2), SomeClass(3)
 
         things = [weakref.WeakSet([obj1, obj2]), obj1, obj2, obj3]
-        result = pickle_depickle(things)
+        result = pickle_depickle(things, protocol=self.protocol)
 
         weakset, depickled1, depickled2, depickled3 = result
 
@@ -652,33 +689,183 @@ class CloudPickleTest(unittest.TestCase):
 
         self.assertEqual(set(weakset), set([depickled1, depickled2]))
 
-    def test_ignoring_whichmodule_exception(self):
-        class FakeModule(object):
-            def __getattr__(self, name):
-                # This throws an exception while looking up within
-                # pickle.whichimodule.
-                raise Exception()
+    def test_faulty_module(self):
+        for module_name in ['_faulty_module', '_missing_module', None]:
+            class FaultyModule(object):
+                def __getattr__(self, name):
+                    # This throws an exception while looking up within
+                    # pickle.whichmodule or getattr(module, name, None)
+                    raise Exception()
 
-        class Foo(object):
-            __module__ = None
+            class Foo(object):
+                __module__ = module_name
 
-            def foo(self):
+                def foo(self):
+                    return "it works!"
+
+            def foo():
                 return "it works!"
 
-        def foo():
-            return "it works!"
+            foo.__module__ = module_name
 
-        foo.__module__ = None
+            sys.modules["_faulty_module"] = FaultyModule()
+            try:
+                # Test whichmodule in save_global.
+                self.assertEqual(pickle_depickle(Foo()).foo(), "it works!")
 
-        sys.modules["_fake_module"] = FakeModule()
-        try:
-            # Test whichmodule in save_global.
-            self.assertEqual(pickle_depickle(Foo()).foo(), "it works!")
+                # Test whichmodule in save_function.
+                cloned = pickle_depickle(foo, protocol=self.protocol)
+                self.assertEqual(cloned(), "it works!")
+            finally:
+                sys.modules.pop("_faulty_module", None)
 
-            # Test whichmodule in save_function.
-            self.assertEqual(pickle_depickle(foo)(), "it works!")
-        finally:
-            sys.modules.pop("_fake_module", None)
+    def test_dynamic_pytest_module(self):
+        # Test case for pull request https://github.com/cloudpipe/cloudpickle/pull/116
+        import py
+
+        def f():
+            s = py.builtin.set([1])
+            return s.pop()
+
+        # some setup is required to allow pytest apimodules to be correctly
+        # serializable.
+        from cloudpickle import CloudPickler
+        CloudPickler.dispatch[type(py.builtin)] = CloudPickler.save_module
+        g = cloudpickle.loads(cloudpickle.dumps(f))
+
+        result = g()
+        self.assertEqual(1, result)
+
+    def test_function_module_name(self):
+        func = lambda x: x
+        cloned = pickle_depickle(func, protocol=self.protocol)
+        self.assertEqual(cloned.__module__, func.__module__)
+
+    def test_function_qualname(self):
+        def func(x):
+            return x
+        # Default __qualname__ attribute (Python 3 only)
+        if hasattr(func, '__qualname__'):
+            cloned = pickle_depickle(func, protocol=self.protocol)
+            self.assertEqual(cloned.__qualname__, func.__qualname__)
+
+        # Mutated __qualname__ attribute
+        func.__qualname__ = '<modifiedlambda>'
+        cloned = pickle_depickle(func, protocol=self.protocol)
+        self.assertEqual(cloned.__qualname__, func.__qualname__)
+
+    def test_namedtuple(self):
+
+        MyTuple = collections.namedtuple('MyTuple', ['a', 'b', 'c'])
+        t = MyTuple(1, 2, 3)
+
+        depickled_t, depickled_MyTuple = pickle_depickle(
+            [t, MyTuple], protocol=self.protocol)
+        self.assertTrue(isinstance(depickled_t, depickled_MyTuple))
+
+        self.assertEqual((depickled_t.a, depickled_t.b, depickled_t.c),
+                         (1, 2, 3))
+        self.assertEqual((depickled_t[0], depickled_t[1], depickled_t[2]),
+                         (1, 2, 3))
+
+        self.assertEqual(depickled_MyTuple.__name__, 'MyTuple')
+        self.assertTrue(issubclass(depickled_MyTuple, tuple))
+
+    def test_builtin_type__new__(self):
+        # Functions occasionally take the __new__ of these types as default
+        # parameters for factories.  For example, on Python 3.3,
+        # `tuple.__new__` is a default value for some methods of namedtuple.
+        for t in list, tuple, set, frozenset, dict, object:
+            cloned = pickle_depickle(t.__new__, protocol=self.protocol)
+            self.assertTrue(cloned is t.__new__)
+
+    def test_interactively_defined_function(self):
+        # Check that callables defined in the __main__ module of a Python
+        # script (or jupyter kernel) can be pickled / unpickled / executed.
+        code = """\
+        from testutils import subprocess_pickle_echo
+
+        CONSTANT = 42
+
+        class Foo(object):
+
+            def method(self, x):
+                return x
+
+        foo = Foo()
+
+        def f0(x):
+            return x ** 2
+
+        def f1():
+            return Foo
+
+        def f2(x):
+            return Foo().method(x)
+
+        def f3():
+            return Foo().method(CONSTANT)
+
+        def f4(x):
+            return foo.method(x)
+
+        cloned = subprocess_pickle_echo(lambda x: x**2, protocol={protocol})
+        assert cloned(3) == 9
+
+        cloned = subprocess_pickle_echo(f0, protocol={protocol})
+        assert cloned(3) == 9
+
+        cloned = subprocess_pickle_echo(Foo, protocol={protocol})
+        assert cloned().method(2) == Foo().method(2)
+
+        cloned = subprocess_pickle_echo(Foo(), protocol={protocol})
+        assert cloned.method(2) == Foo().method(2)
+
+        cloned = subprocess_pickle_echo(f1, protocol={protocol})
+        assert cloned()().method('a') == f1()().method('a')
+
+        cloned = subprocess_pickle_echo(f2, protocol={protocol})
+        assert cloned(2) == f2(2)
+
+        cloned = subprocess_pickle_echo(f3, protocol={protocol})
+        assert cloned() == f3()
+
+        cloned = subprocess_pickle_echo(f4, protocol={protocol})
+        assert cloned(2) == f4(2)
+        """.format(protocol=self.protocol)
+        assert_run_python_script(textwrap.dedent(code))
+
+    @pytest.mark.skipif(sys.version_info >= (3, 0),
+                        reason="hardcoded pickle bytes for 2.7")
+    def test_function_pickle_compat_0_4_0(self):
+        # The result of `cloudpickle.dumps(lambda x: x)` in cloudpickle 0.4.0,
+        # Python 2.7
+        pickled = (b'\x80\x02ccloudpickle.cloudpickle\n_fill_function\nq\x00(c'
+            b'cloudpickle.cloudpickle\n_make_skel_func\nq\x01ccloudpickle.clou'
+            b'dpickle\n_builtin_type\nq\x02U\x08CodeTypeq\x03\x85q\x04Rq\x05(K'
+            b'\x01K\x01K\x01KCU\x04|\x00\x00Sq\x06N\x85q\x07)U\x01xq\x08\x85q'
+            b'\tU\x07<stdin>q\nU\x08<lambda>q\x0bK\x01U\x00q\x0c))tq\rRq\x0eJ'
+            b'\xff\xff\xff\xff}q\x0f\x87q\x10Rq\x11}q\x12N}q\x13NtR.')
+        self.assertEquals(42, cloudpickle.loads(pickled)(42))
+
+    @pytest.mark.skipif(sys.version_info >= (3, 0),
+                        reason="hardcoded pickle bytes for 2.7")
+    def test_function_pickle_compat_0_4_1(self):
+        # The result of `cloudpickle.dumps(lambda x: x)` in cloudpickle 0.4.1,
+        # Python 2.7
+        pickled = (b'\x80\x02ccloudpickle.cloudpickle\n_fill_function\nq\x00(c'
+            b'cloudpickle.cloudpickle\n_make_skel_func\nq\x01ccloudpickle.clou'
+            b'dpickle\n_builtin_type\nq\x02U\x08CodeTypeq\x03\x85q\x04Rq\x05(K'
+            b'\x01K\x01K\x01KCU\x04|\x00\x00Sq\x06N\x85q\x07)U\x01xq\x08\x85q'
+            b'\tU\x07<stdin>q\nU\x08<lambda>q\x0bK\x01U\x00q\x0c))tq\rRq\x0eJ'
+            b'\xff\xff\xff\xff}q\x0f\x87q\x10Rq\x11}q\x12N}q\x13U\x08__main__q'
+            b'\x14NtR.')
+        self.assertEquals(42, cloudpickle.loads(pickled)(42))
+
+
+class Protocol2CloudPickleTest(CloudPickleTest):
+
+    protocol = 2
 
 
 if __name__ == '__main__':
